# WS Environmental Sensor Workout
# Tony Goodhew 19th Sept 2021

# Reads and displays all the values from the sensors fitted to the WaveShare Environment Sensor Module for the Raspberry Pi Pico
# Expects a Waveshare Pi Pico 1.8" Display 160x128 pixels.

from machine import Pin, I2C, SPI, PWM
import framebuf
import utime
import binascii
import math
import struct

# WS160x128 Basic routines
BL = 13
DC = 8
RST = 12
MOSI = 11
SCK = 10
CS = 9

# New Display driver received from WaveShare on 10th July 2021
class LCD_1inch8(framebuf.FrameBuffer):
    def __init__(self):
        self.width = 160
        self.height = 128        
        self.cs = Pin(CS,Pin.OUT)
        self.rst = Pin(RST,Pin.OUT)       
        self.cs(1)
        self.spi = SPI(1)
        self.spi = SPI(1,1000_000)
        self.spi = SPI(1,10000_000,polarity=0, phase=0,sck=Pin(SCK),mosi=Pin(MOSI),miso=None)
        self.dc = Pin(DC,Pin.OUT)
        self.dc(1)
        self.buffer = bytearray(self.height * self.width * 2)
        super().__init__(self.buffer, self.width, self.height, framebuf.RGB565)
        self.init_display()       

        self.WHITE =   0xFFFF
        self.BLACK  =  0x0000
        self.GREEN   =  0x001F
        self.BLUE    =  0xF800
        self.RED   = 0x07E0
        
    def write_cmd(self, cmd):
        self.cs(1)
        self.dc(0)
        self.cs(0)
        self.spi.write(bytearray([cmd]))
        self.cs(1)

    def write_data(self, buf):
        self.cs(1)
        self.dc(1)
        self.cs(0)
        self.spi.write(bytearray([buf]))
        self.cs(1)

    def init_display(self):
        """Initialize display"""  
        self.rst(1)
        self.rst(0)
        self.rst(1)        
        self.write_cmd(0x36);
        self.write_data(0x70);       
        self.write_cmd(0x3A);
        self.write_data(0x05);
         #ST7735R Frame Rate
        self.write_cmd(0xB1);
        self.write_data(0x01);
        self.write_data(0x2C);
        self.write_data(0x2D);

        self.write_cmd(0xB2);
        self.write_data(0x01);
        self.write_data(0x2C);
        self.write_data(0x2D);

        self.write_cmd(0xB3);
        self.write_data(0x01);
        self.write_data(0x2C);
        self.write_data(0x2D);
        self.write_data(0x01);
        self.write_data(0x2C);
        self.write_data(0x2D);

        self.write_cmd(0xB4); #Column inversion
        self.write_data(0x07);

        #ST7735R Power Sequence
        self.write_cmd(0xC0);
        self.write_data(0xA2);
        self.write_data(0x02);
        self.write_data(0x84);
        self.write_cmd(0xC1);
        self.write_data(0xC5);

        self.write_cmd(0xC2);
        self.write_data(0x0A);
        self.write_data(0x00);

        self.write_cmd(0xC3);
        self.write_data(0x8A);
        self.write_data(0x2A);
        self.write_cmd(0xC4);
        self.write_data(0x8A);
        self.write_data(0xEE);

        self.write_cmd(0xC5); #VCOM
        self.write_data(0x0E);

        #ST7735R Gamma Sequence
        self.write_cmd(0xe0);
        self.write_data(0x0f);
        self.write_data(0x1a);
        self.write_data(0x0f);
        self.write_data(0x18);
        self.write_data(0x2f);
        self.write_data(0x28);
        self.write_data(0x20);
        self.write_data(0x22);
        self.write_data(0x1f);
        self.write_data(0x1b);
        self.write_data(0x23);
        self.write_data(0x37);
        self.write_data(0x00);
        self.write_data(0x07);
        self.write_data(0x02);
        self.write_data(0x10);

        self.write_cmd(0xe1);
        self.write_data(0x0f);
        self.write_data(0x1b);
        self.write_data(0x0f);
        self.write_data(0x17);
        self.write_data(0x33);
        self.write_data(0x2c);
        self.write_data(0x29);
        self.write_data(0x2e);
        self.write_data(0x30);
        self.write_data(0x30);
        self.write_data(0x39);
        self.write_data(0x3f);
        self.write_data(0x00);
        self.write_data(0x07);
        self.write_data(0x03);
        self.write_data(0x10);

        self.write_cmd(0xF0); #Enable test command
        self.write_data(0x01);

        self.write_cmd(0xF6); #Disable ram power save mode
        self.write_data(0x00);

            #sleep out
        self.write_cmd(0x11);
        #DEV_Delay_ms(120);

        #Turn on the LCD display
        self.write_cmd(0x29);

    def show(self):
        self.write_cmd(0x2A)
        self.write_data(0x00)
        self.write_data(0x01)
        self.write_data(0x00)
        self.write_data(0xA0)       
        self.write_cmd(0x2B)
        self.write_data(0x00)
        self.write_data(0x02)
        self.write_data(0x00)
        self.write_data(0x82)        
        self.write_cmd(0x2C)        
        self.cs(1)
        self.dc(1)
        self.cs(0)
        self.spi.write(self.buffer)
        self.cs(1)
# ==============End of driver setup==============================
def colour(R,G,B): # Original method
# Get RED value
    rp = int(R*31/255) # range 0 to 31
    if rp < 0: rp = 0
    r = rp *8
# Get Green value - more complicated!
    gp = int(G*63/255) # range 0 - 63
    if gp < 0: gp = 0
    g = 0
    if gp & 1:  g = g + 8192
    if gp & 2:  g = g + 16384
    if gp & 4:  g = g + 32768
    if gp & 8:  g = g + 1
    if gp & 16: g = g + 2
    if gp & 32: g = g + 4
# Get BLUE value       
    bp =int(B*31/255) # range 0 - 31
    if bp < 0: bp = 0
    b = bp *256
    colour = r+g+b
    return colour

def colour2(R,G,B): # Compact method!
    mix1 = ((R&0xF8)*256) + ((G&0xFC)*8) + ((B&0xF8)>>3)
    return  (mix1 & 0xFF) *256  + int((mix1 & 0xFF00) /256) # low nibble first
    
'''
Adjustable font for the WaveShare 1.3" IPS LCD Display Module for Raspberry Pi Pico (240x240)
                         Tony Goodhew 17th Aug 2021
           Modified from code by Les Wright 2021 V 1.1 for Pimoroni Pico Display                  
              https://forums.pimoroni.com/t/pico-display-and-fonts/16194/18
'''
#ASCII Character Set
cmap = ['00000000000000000000000000000000000', #Space
        '00100001000010000100001000000000100', #!
        '01010010100000000000000000000000000', #"
        '01010010101101100000110110101001010', ##
        '00100011111000001110000011111000100', #$
        '11001110010001000100010001001110011', #%
        '01000101001010001000101011001001101', #&
        '10000100001000000000000000000000000', #'
        '00100010001000010000100000100000100', #(
        '00100000100000100001000010001000100', #)
        '00000001001010101110101010010000000', #*
        '00000001000010011111001000010000000', #+
        '000000000000000000000000000000110000100010000', #,
        '00000000000000011111000000000000000', #-
        '00000000000000000000000001100011000', #.
        '00001000010001000100010001000010000', #/
        '01110100011000110101100011000101110', #0
        '00100011000010000100001000010001110', #1
        '01110100010000101110100001000011111', #2
        '01110100010000101110000011000101110', #3
        '00010001100101011111000100001000010', #4
        '11111100001111000001000011000101110', #5
        '01110100001000011110100011000101110', #6
        '11111000010001000100010001000010000', #7
        '01110100011000101110100011000101110', #8
        '01110100011000101111000010000101110', #9
        '00000011000110000000011000110000000', #:
        '01100011000000001100011000010001000', #;
        '00010001000100010000010000010000010', #<
        '00000000001111100000111110000000000', #=
        '01000001000001000001000100010001000', #>
        '01100100100001000100001000000000100', #?
        '01110100010000101101101011010101110', #@
        '00100010101000110001111111000110001', #A
        '11110010010100111110010010100111110', #B
        '01110100011000010000100001000101110', #C
        '11110010010100101001010010100111110', #D
        '11111100001000011100100001000011111', #E
        '11111100001000011100100001000010000', #F
        '01110100011000010111100011000101110', #G
        '10001100011000111111100011000110001', #H
        '01110001000010000100001000010001110', #I
        '00111000100001000010000101001001100', #J
        '10001100101010011000101001001010001', #K
        '10000100001000010000100001000011111', #L
        '10001110111010110101100011000110001', #M
        '10001110011010110011100011000110001', #N
        '01110100011000110001100011000101110', #O
        '11110100011000111110100001000010000', #P
        '01110100011000110001101011001001101', #Q
        '11110100011000111110101001001010001', #R
        '01110100011000001110000011000101110', #S
        '11111001000010000100001000010000100', #T
        '10001100011000110001100011000101110', #U
        '10001100011000101010010100010000100', #V
        '10001100011000110101101011101110001', #W
        '10001100010101000100010101000110001', #X
        '10001100010101000100001000010000100', #Y
        '11111000010001000100010001000011111', #Z
        '01110010000100001000010000100001110', #[
        '10000100000100000100000100000100001', #\
        '00111000010000100001000010000100111', #]
        '00100010101000100000000000000000000', #^
        '00000000000000000000000000000011111', #_
        '11000110001000001000000000000000000', #`
        '00000000000111000001011111000101110', #a
        '10000100001011011001100011100110110', #b
        '00000000000011101000010000100000111', #c
        '00001000010110110011100011001101101', #d
        '00000000000111010001111111000001110', #e
        '00110010010100011110010000100001000', #f
        '000000000001110100011000110001011110000101110', #g
        '10000100001011011001100011000110001', #h
        '00100000000110000100001000010001110', #i
        '0001000000001100001000010000101001001100', #j
        '10000100001001010100110001010010010', #k
        '01100001000010000100001000010001110', #l
        '00000000001101010101101011010110101', #m
        '00000000001011011001100011000110001', #n
        '00000000000111010001100011000101110', #o
        '000000000001110100011000110001111101000010000', #p
        '000000000001110100011000110001011110000100001', #q
        '00000000001011011001100001000010000', #r
        '00000000000111110000011100000111110', #s
        '00100001000111100100001000010000111', #t
        '00000000001000110001100011001101101', #u
        '00000000001000110001100010101000100', #v
        '00000000001000110001101011010101010', #w
        '00000000001000101010001000101010001', #x
        '000000000010001100011000110001011110000101110', #y
        '00000000001111100010001000100011111', #z
        '00010001000010001000001000010000010', #{
        '00100001000010000000001000010000100', #|
        '01000001000010000010001000010001000', #}
        '01000101010001000000000000000000000' #}~
]

def printchar(letter,xpos,ypos,size,charupdate,c):
    origin = xpos
    charval = ord(letter)
    #print(charval)
    index = charval-32 #start code, 32 or space
    #print(index)
    character = cmap[index] #this is our char...
    rows = [character[i:i+5] for i in range(0,len(character),5)]
    #print(rows)
    for row in rows:
        #print(row)
        for bit in row:
            #print(bit)
            if bit == '1':
                LCD.pixel(xpos,ypos,c)
                if size==2:
                    LCD.pixel(xpos,ypos+1,c)
                    LCD.pixel(xpos+1,ypos,c)
                    LCD.pixel(xpos+1,ypos+1,c)
                if size == 3:
                    LCD.pixel(xpos,ypos,c)
                    LCD.pixel(xpos,ypos+1,c)
                    LCD.pixel(xpos,ypos+2,c)
                    LCD.pixel(xpos+1,ypos,c)
                    LCD.pixel(xpos+1,ypos+1,c)
                    LCD.pixel(xpos+1,ypos+2,c)
                    LCD.pixel(xpos+2,ypos,c)
                    LCD.pixel(xpos+2,ypos+1,c)
                    LCD.pixel(xpos+2,ypos+2,c)
                                        
            xpos+=size
        xpos=origin
        ypos+=size
    if charupdate == True:
        LCD.show()
    
def delchar(xpos,ypos,size,delupdate):
    if size == 1:
        charwidth = 5
        charheight = 9
    if size == 2:
        charwidth = 10
        charheight = 18
    if size == 3:
        charwidth = 15
        charheight = 27
    c =colour(0,0,0) # Colour of background
    LCD.fill_rect(xpos,ypos,charwidth,charheight,c) #xywh
    if delupdate == True:
        LCD.show()

def printstring(string,xpos,ypos,size,charupdate,strupdate,c):   
    if size == 1:
        spacing = 8
    if size == 2:
        spacing = 14
    if size == 3:
        spacing = 18
    for i in string:
        printchar(i,xpos,ypos,size,charupdate,c)
        xpos+=spacing
    if strupdate == True:
        LCD.show()
        
def centrestring(string,ypos,size,charupdate,strupdate,c):
    if size == 1:
        spacing = 8
    if size == 2:
        spacing = 14
    if size == 3:
        spacing = 18
    used = len(string)*spacing
    xpos = int((159-used)/2)
    printstring(string,xpos,ypos,size,charupdate,strupdate,c)
# =============End of Characters section ===============
# =============== BME280 ==============
I2C_ADDR = 0x76

digT = []
digP = []
digH = []

t_fine = 0.0

class BME280:
	def __init__(self, address = I2C_ADDR):
		self.i2c = I2C(0, scl=Pin(21), sda=Pin(20), freq=100000)
		self.address = address
		
		self.calib = []
		self.osrs_t = 1			#Temperature oversampling x 1
		self.osrs_p = 1			#Pressure oversampling x 1
		self.osrs_h = 1			#Humidity oversampling x 1
		self.mode   = 3			#Normal self.mode
		self.t_sb   = 5			#Tstandby 1000ms
		self.filter = 0			#self.filter off
		self.spi3w_en = 0			#3-wire SPI Disable

		ctrl_meas_reg = (self.osrs_t << 5) | (self.osrs_p << 2) | self.mode
		config_reg    = (self.t_sb << 5) | (self.filter << 2) | self.spi3w_en
		ctrl_hum_reg  = self.osrs_h

		self.writeReg(0xF2, ctrl_hum_reg)
		self.writeReg(0xF4, ctrl_meas_reg)
		self.writeReg(0xF5, config_reg)
			
	def writeReg(self, cmd, val):
		self.i2c.writeto_mem(int(self.address), int(cmd), bytes([int(val)]))
        
	def get_calib_param(self):
		for i in range (0x88,0x88+24):
			rdate = self.i2c.readfrom_mem(int(self.address), int(i), 1)
			self.calib.append(rdate[0])
		rdate2 = self.i2c.readfrom_mem(int(self.address), int(0xA1), 1)
		self.calib.append(rdate2[0])
		for i in range (0xE1,0xE1+7):
			rdate3 = self.i2c.readfrom_mem(int(self.address), int(i), 1)
			self.calib.append(rdate3[0])

		digT.append((self.calib[1] << 8) | self.calib[0])
		digT.append((self.calib[3] << 8) | self.calib[2])
		digT.append((self.calib[5] << 8) | self.calib[4])
		digP.append((self.calib[7] << 8) | self.calib[6])
		digP.append((self.calib[9] << 8) | self.calib[8])
		digP.append((self.calib[11]<< 8) | self.calib[10])
		digP.append((self.calib[13]<< 8) | self.calib[12])
		digP.append((self.calib[15]<< 8) | self.calib[14])
		digP.append((self.calib[17]<< 8) | self.calib[16])
		digP.append((self.calib[19]<< 8) | self.calib[18])
		digP.append((self.calib[21]<< 8) | self.calib[20])
		digP.append((self.calib[23]<< 8) | self.calib[22])
		digH.append( self.calib[24] )
		digH.append((self.calib[26]<< 8) | self.calib[25])
		digH.append( self.calib[27] )
		digH.append((self.calib[28]<< 4) | (0x0F & self.calib[29]))
		digH.append((self.calib[30]<< 4) | ((self.calib[29] >> 4) & 0x0F))
		digH.append( self.calib[31] )
		
		for i in range(1,2):
			if digT[i] & 0x8000:
				digT[i] = (-digT[i] ^ 0xFFFF) + 1

		for i in range(1,8):
			if digP[i] & 0x8000:
				digP[i] = (-digP[i] ^ 0xFFFF) + 1

		for i in range(0,6):
			if digH[i] & 0x8000:
				digH[i] = (-digH[i] ^ 0xFFFF) + 1  

	def readData(self):
		data = []
		for i in range (0xF7, 0xF7+8):
			rdate = self.i2c.readfrom_mem(int(self.address), int(i), 1)
			data.append(rdate[0])
		pres_raw = (data[0] << 12) | (data[1] << 4) | (data[2] >> 4)
		temp_raw = (data[3] << 12) | (data[4] << 4) | (data[5] >> 4)
		hum_raw  = (data[6] << 8)  |  data[7]
		
		pressure = self.compensate_P(pres_raw)
		temperature = self.compensate_T(temp_raw)
		var_h = self.compensate_H(hum_raw)
		# print "pressure : %7.2f hPa" % (pressure/100)
		# print "temp : %-6.2f ℃" % (temperature) 
		# print "hum : %6.2f ％" % (var_h)
		return [pressure, temperature, var_h]

	def compensate_P(self, adc_P):
		global  t_fine
		pressure = 0.0
		
		v1 = (t_fine / 2.0) - 64000.0
		v2 = (((v1 / 4.0) * (v1 / 4.0)) / 2048) * digP[5]
		v2 = v2 + ((v1 * digP[4]) * 2.0)
		v2 = (v2 / 4.0) + (digP[3] * 65536.0)
		v1 = (((digP[2] * (((v1 / 4.0) * (v1 / 4.0)) / 8192)) / 8)  + ((digP[1] * v1) / 2.0)) / 262144
		v1 = ((32768 + v1) * digP[0]) / 32768
		
		if v1 == 0:
			return 0
		pressure = ((1048576 - adc_P) - (v2 / 4096)) * 3125
		if pressure < 0x80000000:
			pressure = (pressure * 2.0) / v1
		else:
			pressure = (pressure / v1) * 2
		v1 = (digP[8] * (((pressure / 8.0) * (pressure / 8.0)) / 8192.0)) / 4096
		v2 = ((pressure / 4.0) * digP[7]) / 8192.0
		pressure = pressure + ((v1 + v2 + digP[6]) / 16.0)  
		return (pressure/100)
		# print "pressure : %7.2f hPa" % (pressure/100)

	def compensate_T(self, adc_T):
		global t_fine
		v1 = (adc_T / 16384.0 - digT[0] / 1024.0) * digT[1]
		v2 = (adc_T / 131072.0 - digT[0] / 8192.0) * (adc_T / 131072.0 - digT[0] / 8192.0) * digT[2]
		t_fine = v1 + v2
		temperature = t_fine / 5120.0
		# print "temp : %-6.2f ℃" % (temperature) 
		return temperature

	def compensate_H(self, adc_H):
		global t_fine
		var_h = t_fine - 76800.0
		if var_h != 0:
			var_h = (adc_H - (digH[3] * 64.0 + digH[4]/16384.0 * var_h)) * (digH[1] / 65536.0 * (1.0 + digH[5] / 67108864.0 * var_h * (1.0 + digH[2] / 67108864.0 * var_h)))
		else:
			return 0
		var_h = var_h * (1.0 - digH[0] * var_h / 524288.0)
		if var_h > 100.0:
			var_h = 100.0
		elif var_h < 0.0:
			var_h = 0.0
		# print "hum : %6.2f ％" % (var_h)
		return var_h
# =========== End of BME 280 ================

# ==========Start of SPG40 =============
SGP40_CMD_FEATURE_SET = [0x20, 0x2F]
SGP40_CMD_MEASURE_TEST = [0X28,0X0E]
SGP40_CMD_SOFT_RESET = [0X00,0X06]
SGP40_CMD_HEATER_OFF = [0X36,0X15]
SGP40_CMD_MEASURE_RAW = [0X26,0X0F]
# SGP40_CMD_GET_SERIAL_ID = [0X36,0X82] #DATASHEET is not written ,but Sensirion have


CRC_TABLE = [
        0, 49, 98, 83, 196, 245, 166, 151, 185, 136, 219, 234, 125, 76, 31, 46,
        67, 114, 33, 16, 135, 182, 229, 212, 250, 203, 152, 169, 62, 15, 92, 109,
        134, 183, 228, 213, 66, 115, 32, 17, 63, 14, 93, 108, 251, 202, 153, 168,
        197, 244, 167, 150, 1, 48, 99, 82, 124, 77, 30, 47, 184, 137, 218, 235,
        61, 12, 95, 110, 249, 200, 155, 170, 132, 181, 230, 215, 64, 113, 34, 19,
        126, 79, 28, 45, 186, 139, 216, 233, 199, 246, 165, 148, 3, 50, 97, 80,
        187, 138, 217, 232, 127, 78, 29, 44, 2, 51, 96, 81, 198, 247, 164, 149,
        248, 201, 154, 171, 60, 13, 94, 111, 65, 112, 35, 18, 133, 180, 231, 214,
        122, 75, 24, 41, 190, 143, 220, 237, 195, 242, 161, 144, 7, 54, 101, 84,
        57, 8, 91, 106, 253, 204, 159, 174, 128, 177, 226, 211, 68, 117, 38, 23,
        252, 205, 158, 175, 56, 9, 90, 107, 69, 116, 39, 22, 129, 176, 227, 210,
        191, 142, 221, 236, 123, 74, 25, 40, 6, 55, 100, 85, 194, 243, 160, 145,
        71, 118, 37, 20, 131, 178, 225, 208, 254, 207, 156, 173, 58, 11, 88, 105,
        4, 53, 102, 87, 192, 241, 162, 147, 189, 140, 223, 238, 121, 72, 27, 42,
        193, 240, 163, 146, 5, 52, 103, 86, 120, 73, 26, 43, 188, 141, 222, 239,
        130, 179, 224, 209, 70, 119, 36, 21, 59, 10, 89, 104, 255, 206, 157, 172
        ]

#Without_humidity_compensation
#sgp40_measure_raw + 2*humi + CRC + 2*temp + CRC
WITHOUT_HUM_COMP = [0X26, 0X0F, 0X80, 0X00, 0XA2, 0X66, 0X66, 0X93] # default Temperature=25 Humidity=50
WITH_HUM_COMP = [0x26, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00] #Manual input

ADDR = 0x59

class SGP40:
    def __init__(self, address=ADDR):
        self.i2c = I2C(0, scl=Pin(21), sda=Pin(20), freq=100000)
        self.address = address
        
        # feature set 0x3220
        self.write(SGP40_CMD_FEATURE_SET)    
        utime.sleep(0.25)
        Rbuf = self.Read() 
        # print('feature set:  :%#x'% ((int(Rbuf[0]) << 8) | Rbuf[1]))
        if ((int(Rbuf[0]) << 8) | Rbuf[1]) != 0x3220:
            raise RunutimeError("Self test failed")
        
        # Self Test 0xD400      
        self.write(SGP40_CMD_MEASURE_TEST)    
        utime.sleep(0.25)
        Rbuf = self.Read()
        # print('Self Test  :%#x'% ((int(Rbuf[0]) << 8) | Rbuf[1]))
        if ((int(Rbuf[0]) << 8) | Rbuf[1]) != 0xD400: #0x4B00 is failed,0xD400 pass
            raise RunutimeError("Self test failed")
            
        # reset
        # self.write(SGP40_CMD_SOFT_RESET)

    def Read(self):
        rdate = self.i2c.readfrom_mem(int(self.address), 0, 3)
        return rdate
    
    def write(self, cmd):
		self.i2c.writeto_mem(int(self.address), int(cmd[0]), bytes([int(cmd[1])]))
        
    def write_block(self, cmd):
        self.i2c.writeto_mem(int(self.address), int(cmd[0]), bytes(cmd[1:8]))
        
    def raw(self):
        """The raw gas value"""
        # recycle a single buffer
        self.write_block(WITHOUT_HUM_COMP)
        utime.sleep(0.25)
        Rbuf = self.Read()
        return ((int(Rbuf[0]) << 8) | Rbuf[1])
        
    def measureRaw(self, temperature, humidity):
        # 2*humi + CRC
        paramh = struct.pack(">H", math.ceil(humidity * 0xffff / 100))
        crch = self.__crc(paramh[0], paramh[1])
        
        # 2*temp + CRC
        paramt = struct.pack(">H", math.ceil((temperature + 45) * 0xffff / 175))
        crct = self.__crc(paramt[0], paramt[1])
        
        WITH_HUM_COMP[2:3] = list(paramh)
        WITH_HUM_COMP[4] = int(crch)
        WITH_HUM_COMP[5:6] = list(paramt)
        WITH_HUM_COMP[7] = int(crct)
        self.write_block(WITH_HUM_COMP)
        
        utime.sleep(0.5)
        Rbuf = self.Read()
        return ((int(Rbuf[0]) << 8) | Rbuf[1])
        
    def __crc(self, msb, lsb):
        crc = 0xff
        crc ^= msb
        crc = CRC_TABLE[crc]
        if lsb is not None:
            crc ^= lsb
            crc = CRC_TABLE[crc]
        return crc
# ========= End of SPG40 ==========
# === Start of TSL2591 Lux ==============
ADDR                = (0x29)

COMMAND_BIT         = (0xA0)
#Register (0x00)
ENABLE_REGISTER     = (0x00)
ENABLE_POWERON      = (0x01)
ENABLE_POWEROFF     = (0x00)
ENABLE_AEN          = (0x02)
ENABLE_AIEN         = (0x10)
ENABLE_SAI          = (0x40)
ENABLE_NPIEN        = (0x80)

CONTROL_REGISTER    = (0x01)
SRESET              = (0x80)

AILTL_REGISTER      = (0x04)
AILTH_REGISTER      = (0x05)
AIHTL_REGISTER      = (0x06)
AIHTH_REGISTER      = (0x07)
NPAILTL_REGISTER    = (0x08)
NPAILTH_REGISTER    = (0x09)
NPAIHTL_REGISTER    = (0x0A)
NPAIHTH_REGISTER    = (0x0B)

PERSIST_REGISTER    = (0x0C)
# Bits 3:0
# 0000          Every ALS cycle generates an interrupt
# 0001          Any value outside of threshold range
# 0010          2 consecutive values out of range
# 0011          3 consecutive values out of range
# 0100          5 consecutive values out of range
# 0101          10 consecutive values out of range
# 0110          15 consecutive values out of range
# 0111          20 consecutive values out of range
# 1000          25 consecutive values out of range
# 1001          30 consecutive values out of range
# 1010          35 consecutive values out of range
# 1011          40 consecutive values out of range
# 1100          45 consecutive values out of range
# 1101          50 consecutive values out of range
# 1110          55 consecutive values out of range
# 1111          60 consecutive values out of range

ID_REGISTER         = (0x12)

STATUS_REGISTER     = (0x13)#read only

CHAN0_LOW           = (0x14)
CHAN0_HIGH          = (0x15)
CHAN1_LOW           = (0x16)
CHAN1_HIGH          = (0x17)

LUX_DF = 408.0
LUX_COEFB = 1.64
LUX_COEFC = 0.59
LUX_COEFD = 0.86

#AGAIN
LOW_AGAIN           = (0X00)#Low gain (1x)
MEDIUM_AGAIN        = (0X10)#Medium gain (25x)
HIGH_AGAIN          = (0X20)#High gain (428x)
MAX_AGAIN           = (0x30)#Max gain (9876x)

#Autime
Autime_100MS         = (0x00)#100 millis #MAX COUNT 36863 
Autime_200MS         = (0x01)#200 millis #MAX COUNT 65535 
Autime_300MS         = (0x02)#300 millis
Autime_400MS         = (0x03)#400 millis
Autime_500MS         = (0x04)#500 millis
Autime_600MS         = (0x05)#600 millis

MAX_COUNT_100MS     = (36863) # 0x8FFF
MAX_COUNT           = (65535) # 0xFFFF

class TSL2591:
    def __init__(self, address=ADDR):
		self.i2c = I2C(0, scl=Pin(21), sda=Pin(20), freq=100000)
		self.address = address

		self.ID = self.Read_Byte(ID_REGISTER)
		if(self.ID != 0x50):
			print("ID = ",self.ID)
		# self.Write_Byte(ENABLE_REGISTER, ENABLE_POWERON | ENABLE_AEN)
		self.Write_Byte(ENABLE_REGISTER, ENABLE_AIEN | ENABLE_POWERON | ENABLE_AEN | ENABLE_NPIEN)
		self.Integralutime = Autime_200MS
		self.Gain = MEDIUM_AGAIN
		self.Write_Byte(CONTROL_REGISTER, self.Integralutime | self.Gain)
		self.Write_Byte(PERSIST_REGISTER, 0x01)
		# self.Write_Byte(0xE7, 0x13)
		autime = 100.0 * self.Integralutime + 100.0
		again = 1.0
		if self.Gain == MEDIUM_AGAIN:
			again = 25.0
		elif self.Gain == HIGH_AGAIN:
			again = 428.0
		elif self.Gain == MAX_AGAIN:
			again = 9876.0
		self.Cpl = (autime * again) / LUX_DF

    def Read_Byte(self, cmd):
        cmd = (COMMAND_BIT | cmd) & 0xFF
        rdate = self.i2c.readfrom_mem(int(self.address), int(cmd), 1)
        return rdate[0]

    def Write_Byte(self, cmd, val):
        cmd = (COMMAND_BIT | cmd) & 0xFF
        self.i2c.writeto_mem(int(self.address), int(cmd), bytes([int(val)]))
        
    def Read_2Channel(self):
		CH0L = self.Read_Byte(CHAN0_LOW)
		CH0H = self.Read_Byte(CHAN0_LOW + 1)
		CH1L = self.Read_Byte(CHAN0_LOW + 2)
		CH1H = self.Read_Byte(CHAN0_LOW + 3)
		full = (CH0H << 8)|CH0L
		ir = (CH1H << 8)|CH1L
		return full,ir

    def Lux(self):
		status = self.Read_Byte(0x13)
		if(status & 0x10):
			# print ('soft goto interrupt')
			self.Write_Byte(0xE7, 0x13)
		
		full, ir = self.Read_2Channel()
		if full == 0xFFFF or ir == 0xFFFF:
			raise RunutimeError('Numerical overflow!')

		# lux1 = (full - (LUX_COEFB * ir)) / self.Cpl
		# lux2 = ((LUX_COEFC * full) - (LUX_COEFD * ir)) / self.Cpl
		# return max(int(lux1), int(lux2))
		lux = ((full-ir) * (1.00 - (ir/full))) / self.Cpl
		# lux = (full-ir)/ self.Cpl
		return lux,ir                         # I added the IR to return - Correct?  

    def SET_LuxInterrupt(self, SET_LOW, SET_HIGH):
		full, ir = self.Read_2Channel()
		set0dataL = int(SET_LOW * self.Cpl + ir)
		set0dataH = int(SET_HIGH * self.Cpl + ir)

		self.Write_Byte(AILTL_REGISTER, set0dataL & 0xFF)
		self.Write_Byte(AILTH_REGISTER, set0dataL >> 8)

		self.Write_Byte(AIHTL_REGISTER, set0dataH & 0xFF)
		self.Write_Byte(AIHTH_REGISTER, set0dataH >> 8)
# ============ End of Lux sensor ==============
# ========= Start LTR390 UV sensor driver =============
ADDR  = (0X53)

LTR390_MAIN_CTRL = (0x00)  # Main control register
LTR390_MEAS_RATE = (0x04)  # Resolution and data rate
LTR390_GAIN = (0x05)  # ALS and UVS gain range
LTR390_PART_ID = (0x06)  # Part id/revision register
LTR390_MAIN_STATUS = (0x07)  # Main status register
LTR390_ALSDATA = (0x0D)  # ALS data lowest byte, 3 byte
LTR390_UVSDATA = (0x10)  # UVS data lowest byte, 3 byte
LTR390_INT_CFG = (0x19)  # Interrupt configuration
LTR390_INT_PST = (0x1A)  # Interrupt persistance config
LTR390_THRESH_UP = (0x21)  # Upper threshold, low byte, 3 byte
LTR390_THRESH_LOW = (0x24)  # Lower threshold, low byte, 3 byte

#ALS/UVS measurement resolution, Gain setting, measurement rate
RESOLUTION_20BIT_utime400MS = (0X00)
RESOLUTION_19BIT_utime200MS = (0X10)
RESOLUTION_18BIT_utime100MS = (0X20)#default
RESOLUTION_17BIT_utime50MS  = (0x3)
RESOLUTION_16BIT_utime25MS  = (0x40)
RESOLUTION_13BIT_utime12_5MS  = (0x50)
RATE_25MS = (0x0)
RATE_50MS = (0x1)
RATE_100MS = (0x2)# default
RATE_200MS = (0x3)
RATE_500MS = (0x4)
RATE_1000MS = (0x5)
RATE_2000MS = (0x6)

# measurement Gain Range.
GAIN_1  = (0x0)
GAIN_3  = (0x1)# default
GAIN_6 = (0x2)
GAIN_9 = (0x3)
GAIN_18 = (0x4)

class LTR390:
    def __init__(self, address=ADDR):
        self.i2c = I2C(0, scl=Pin(21), sda=Pin(20), freq=100000)
        self.address = address

        self.ID = self.Read_Byte(LTR390_PART_ID)
        # print("ID = %#x" %self.ID)
        if(self.ID != 0xB2):
            print("read ID error!,Check the hardware...")
            return

        self.Write_Byte(LTR390_MAIN_CTRL, 0x0A) #  UVS in Active Mode
        self.Write_Byte(LTR390_MEAS_RATE, RESOLUTION_20BIT_utime400MS | RATE_2000MS) #  Resolution=18bits, Meas Rate = 100ms
        self.Write_Byte(LTR390_GAIN, GAIN_3) #  Gain Range=3.
        # self.Write_Byte(LTR390_INT_CFG, 0x34) # UVS_INT_EN=1, Command=0x34
        # self.Write_Byte(LTR390_GAIN, GAIN_3) #  Resolution=18bits, Meas Rate = 100ms
        
    def Read_Byte(self, cmd):
        rdate = self.i2c.readfrom_mem(int(self.address), int(cmd), 1)
        return rdate[0]

    def Write_Byte(self, cmd, val):
        self.i2c.writeto_mem(int(self.address), int(cmd), bytes([int(val)]))
        
    def UVS(self):
        # self.Write_Byte(LTR390_MAIN_CTRL, 0x0A) #  UVS in Active Mode
        Data1 = self.Read_Byte(LTR390_UVSDATA)
        Data2 = self.Read_Byte(LTR390_UVSDATA + 1)
        Data3 = self.Read_Byte(LTR390_UVSDATA + 2)
        uv =  (Data3 << 16)| (Data2 << 8) | Data1
        # UVS = Data3*65536+Data2*256+Data1
        # print("UVS = ", UVS)
        return uv
# ========= End LTR390 UV sensor driver =============
# ======== Start ICM20948 9-DOF Sensor

Gyro  = [0,0,0]
Accel = [0,0,0]
Mag   = [0,0,0]
pitch = 0.0
roll  = 0.0
yaw   = 0.0
pu8data=[0,0,0,0,0,0,0,0]
U8tempX=[0,0,0,0,0,0,0,0,0]
U8tempY=[0,0,0,0,0,0,0,0,0]
U8tempZ=[0,0,0,0,0,0,0,0,0]
GyroOffset=[0,0,0]
Ki = 1.0
Kp = 4.50
q0 = 1.0
q1=q2=q3=0.0
angles=[0.0,0.0,0.0]
true                                 =0x01
false                                =0x00
# define ICM-20948 Device I2C address
I2C_ADD_ICM20948                     = 0x68
I2C_ADD_ICM20948_AK09916             = 0x0C
I2C_ADD_ICM20948_AK09916_READ        = 0x80
I2C_ADD_ICM20948_AK09916_WRITE       = 0x00
# define ICM-20948 Register
# user bank 0 register
REG_ADD_WIA                          = 0x00
REG_VAL_WIA                          = 0xEA
REG_ADD_USER_CTRL                    = 0x03
REG_VAL_BIT_DMP_EN                   = 0x80
REG_VAL_BIT_FIFO_EN                  = 0x40
REG_VAL_BIT_I2C_MST_EN               = 0x20
REG_VAL_BIT_I2C_IF_DIS               = 0x10
REG_VAL_BIT_DMP_RST                  = 0x08
REG_VAL_BIT_DIAMOND_DMP_RST          = 0x04
REG_ADD_PWR_MIGMT_1                  = 0x06
REG_VAL_ALL_RGE_RESET                = 0x80
REG_VAL_RUN_MODE                     = 0x01 # Non low-power mode
REG_ADD_LP_CONFIG                    = 0x05
REG_ADD_PWR_MGMT_1                   = 0x06
REG_ADD_PWR_MGMT_2                   = 0x07
REG_ADD_ACCEL_XOUT_H                 = 0x2D
REG_ADD_ACCEL_XOUT_L                 = 0x2E
REG_ADD_ACCEL_YOUT_H                 = 0x2F
REG_ADD_ACCEL_YOUT_L                 = 0x30
REG_ADD_ACCEL_ZOUT_H                 = 0x31
REG_ADD_ACCEL_ZOUT_L                 = 0x32
REG_ADD_GYRO_XOUT_H                  = 0x33
REG_ADD_GYRO_XOUT_L                  = 0x34
REG_ADD_GYRO_YOUT_H                  = 0x35
REG_ADD_GYRO_YOUT_L                  = 0x36
REG_ADD_GYRO_ZOUT_H                  = 0x37
REG_ADD_GYRO_ZOUT_L                  = 0x38
REG_ADD_EXT_SENS_DATA_00             = 0x3B
REG_ADD_REG_BANK_SEL                 = 0x7F
REG_VAL_REG_BANK_0                   = 0x00
REG_VAL_REG_BANK_1                   = 0x10
REG_VAL_REG_BANK_2                   = 0x20
REG_VAL_REG_BANK_3                   = 0x30

# user bank 1 register
# user bank 2 register
REG_ADD_GYRO_SMPLRT_DIV              = 0x00
REG_ADD_GYRO_CONFIG_1                = 0x01
REG_VAL_BIT_GYRO_DLPCFG_2            = 0x10  # bit[5:3]
REG_VAL_BIT_GYRO_DLPCFG_4            = 0x20  # bit[5:3]
REG_VAL_BIT_GYRO_DLPCFG_6            = 0x30  # bit[5:3]
REG_VAL_BIT_GYRO_FS_250DPS           = 0x00  # bit[2:1]
REG_VAL_BIT_GYRO_FS_500DPS           = 0x02  # bit[2:1]
REG_VAL_BIT_GYRO_FS_1000DPS          = 0x04  # bit[2:1]
REG_VAL_BIT_GYRO_FS_2000DPS          = 0x06  # bit[2:1]
REG_VAL_BIT_GYRO_DLPF                = 0x01  # bit[0]
REG_ADD_ACCEL_SMPLRT_DIV_2           = 0x11
REG_ADD_ACCEL_CONFIG                 = 0x14
REG_VAL_BIT_ACCEL_DLPCFG_2           = 0x10  # bit[5:3]
REG_VAL_BIT_ACCEL_DLPCFG_4           = 0x20  # bit[5:3]
REG_VAL_BIT_ACCEL_DLPCFG_6           = 0x30  # bit[5:3]
REG_VAL_BIT_ACCEL_FS_2g              = 0x00  # bit[2:1]
REG_VAL_BIT_ACCEL_FS_4g              = 0x02  # bit[2:1]
REG_VAL_BIT_ACCEL_FS_8g              = 0x04  # bit[2:1]
REG_VAL_BIT_ACCEL_FS_16g             = 0x06  # bit[2:1]
REG_VAL_BIT_ACCEL_DLPF               = 0x01  # bit[0]

# user bank 3 register
REG_ADD_I2C_SLV0_ADDR                = 0x03
REG_ADD_I2C_SLV0_REG                 = 0x04
REG_ADD_I2C_SLV0_CTRL                = 0x05
REG_VAL_BIT_SLV0_EN                  = 0x80
REG_VAL_BIT_MASK_LEN                 = 0x07
REG_ADD_I2C_SLV0_DO                  = 0x06
REG_ADD_I2C_SLV1_ADDR                = 0x07
REG_ADD_I2C_SLV1_REG                 = 0x08
REG_ADD_I2C_SLV1_CTRL                = 0x09
REG_ADD_I2C_SLV1_DO                  = 0x0A

# define ICM-20948 Register  end

# define ICM-20948 MAG Register
REG_ADD_MAG_WIA1                     = 0x00
REG_VAL_MAG_WIA1                     = 0x48
REG_ADD_MAG_WIA2                     = 0x01
REG_VAL_MAG_WIA2                     = 0x09
REG_ADD_MAG_ST2                      = 0x10
REG_ADD_MAG_DATA                     = 0x11
REG_ADD_MAG_CNTL2                    = 0x31
REG_VAL_MAG_MODE_PD                  = 0x00
REG_VAL_MAG_MODE_SM                  = 0x01
REG_VAL_MAG_MODE_10HZ                = 0x02
REG_VAL_MAG_MODE_20HZ                = 0x04
REG_VAL_MAG_MODE_50HZ                = 0x05
REG_VAL_MAG_MODE_100HZ               = 0x08
REG_VAL_MAG_MODE_ST                  = 0x10
# define ICM-20948 MAG Register  end

MAG_DATA_LEN                         =6

class ICM20948(object):
  def __init__(self,address=I2C_ADD_ICM20948):
    self._address = address
    self._bus = I2C(0, scl=Pin(21), sda=Pin(20), freq=100000)
    bRet=self.icm20948Check()             #Initialization of the device multiple times after power on will result in a return error
    utime.sleep(0.5)                       #We can skip this detection by delaying it by 500 milliseconds
    # user bank 0 register 
    self._write_byte( REG_ADD_REG_BANK_SEL , REG_VAL_REG_BANK_0)
    self._write_byte( REG_ADD_PWR_MIGMT_1 , REG_VAL_ALL_RGE_RESET)
    utime.sleep(0.1)
    self._write_byte( REG_ADD_PWR_MIGMT_1 , REG_VAL_RUN_MODE)  
    #user bank 2 register
    self._write_byte( REG_ADD_REG_BANK_SEL , REG_VAL_REG_BANK_2)
    self._write_byte( REG_ADD_GYRO_SMPLRT_DIV , 0x07)
    self._write_byte( REG_ADD_GYRO_CONFIG_1 , REG_VAL_BIT_GYRO_DLPCFG_6 | REG_VAL_BIT_GYRO_FS_1000DPS | REG_VAL_BIT_GYRO_DLPF)
    self._write_byte( REG_ADD_ACCEL_SMPLRT_DIV_2 ,  0x07)
    self._write_byte( REG_ADD_ACCEL_CONFIG , REG_VAL_BIT_ACCEL_DLPCFG_6 | REG_VAL_BIT_ACCEL_FS_2g | REG_VAL_BIT_ACCEL_DLPF)
    #user bank 0 register
    self._write_byte( REG_ADD_REG_BANK_SEL , REG_VAL_REG_BANK_0) 
    utime.sleep(0.1)
    self.GyroOffset()
    self.MagCheck()
    self.WriteSecondary( I2C_ADD_ICM20948_AK09916|I2C_ADD_ICM20948_AK09916_WRITE,REG_ADD_MAG_CNTL2, REG_VAL_MAG_MODE_20HZ)
	
  def Gyro_Accel_Read(self):
    self._write_byte( REG_ADD_REG_BANK_SEL , REG_VAL_REG_BANK_0)
    data =self._read_block(REG_ADD_ACCEL_XOUT_H, 12)
    self._write_byte( REG_ADD_REG_BANK_SEL , REG_VAL_REG_BANK_2)
    Accel[0] = (data[0]<<8)|data[1]
    Accel[1] = (data[2]<<8)|data[3]
    Accel[2] = (data[4]<<8)|data[5]
    Gyro[0]  = ((data[6]<<8)|data[7]) - GyroOffset[0]
    Gyro[1]  = ((data[8]<<8)|data[9]) - GyroOffset[1]
    Gyro[2]  = ((data[10]<<8)|data[11]) - GyroOffset[2]
    if Accel[0]>=32767:             #Solve the problem that Python shift will not overflow
      Accel[0]=Accel[0]-65535
    elif Accel[0]<=-32767:
      Accel[0]=Accel[0]+65535
    if Accel[1]>=32767:
      Accel[1]=Accel[1]-65535
    elif Accel[1]<=-32767:
      Accel[1]=Accel[1]+65535
    if Accel[2]>=32767:
      Accel[2]=Accel[2]-65535
    elif Accel[2]<=-32767:
      Accel[2]=Accel[2]+65535
    if Gyro[0]>=32767:
      Gyro[0]=Gyro[0]-65535
    elif Gyro[0]<=-32767:
      Gyro[0]=Gyro[0]+65535
    if Gyro[1]>=32767:
      Gyro[1]=Gyro[1]-65535
    elif Gyro[1]<=-32767:
      Gyro[1]=Gyro[1]+65535
    if Gyro[2]>=32767:
      Gyro[2]=Gyro[2]-65535
    elif Gyro[2]<=-32767:
      Gyro[2]=Gyro[2]+65535
	  
  def MagRead(self):
    counter=20
    while(counter>0):
      utime.sleep(0.01)
      self.icm20948ReadSecondary( I2C_ADD_ICM20948_AK09916|I2C_ADD_ICM20948_AK09916_READ , REG_ADD_MAG_ST2, 1)
      if ((pu8data[0] & 0x01)!= 0):
        break
      counter-=1
    if counter!=0:
      for i in range(0,8):
        self.icm20948ReadSecondary( I2C_ADD_ICM20948_AK09916|I2C_ADD_ICM20948_AK09916_READ , REG_ADD_MAG_DATA , MAG_DATA_LEN)
        U8tempX[i] = (pu8data[1]<<8)|pu8data[0]
        U8tempY[i] = (pu8data[3]<<8)|pu8data[2]
        U8tempZ[i] = (pu8data[5]<<8)|pu8data[4]
      Mag[0]=(U8tempX[0]+U8tempX[1]+U8tempX[2]+U8tempX[3]+U8tempX[4]+U8tempX[5]+U8tempX[6]+U8tempX[7])/8
      Mag[1]=-(U8tempY[0]+U8tempY[1]+U8tempY[2]+U8tempY[3]+U8tempY[4]+U8tempY[5]+U8tempY[6]+U8tempY[7])/8
      Mag[2]=-(U8tempZ[0]+U8tempZ[1]+U8tempZ[2]+U8tempZ[3]+U8tempZ[4]+U8tempZ[5]+U8tempZ[6]+U8tempZ[7])/8
    if Mag[0]>=32767:            #Solve the problem that Python shift will not overflow
      Mag[0]=Mag[0]-65535
    elif Mag[0]<=-32767:
      Mag[0]=Mag[0]+65535
    if Mag[1]>=32767:
      Mag[1]=Mag[1]-65535
    elif Mag[1]<=-32767:
      Mag[1]=Mag[1]+65535
    if Mag[2]>=32767:
      Mag[2]=Mag[2]-65535
    elif Mag[2]<=-32767:
      Mag[2]=Mag[2]+65535
	  
  def icm20948ReadSecondary(self,u8I2CAddr,u8RegAddr,u8Len):
    u8Temp=0
    self._write_byte( REG_ADD_REG_BANK_SEL,  REG_VAL_REG_BANK_3) #swtich bank3
    self._write_byte( REG_ADD_I2C_SLV0_ADDR, u8I2CAddr)
    self._write_byte( REG_ADD_I2C_SLV0_REG,  u8RegAddr)
    self._write_byte( REG_ADD_I2C_SLV0_CTRL, REG_VAL_BIT_SLV0_EN|u8Len)

    self._write_byte( REG_ADD_REG_BANK_SEL, REG_VAL_REG_BANK_0) #swtich bank0
    
    u8Temp = self._read_byte(REG_ADD_USER_CTRL)
    u8Temp |= REG_VAL_BIT_I2C_MST_EN
    self._write_byte( REG_ADD_USER_CTRL, u8Temp)
    utime.sleep(0.01)
    u8Temp &= ~REG_VAL_BIT_I2C_MST_EN
    self._write_byte( REG_ADD_USER_CTRL, u8Temp)
    
    for i in range(0,u8Len):
      pu8data[i]= self._read_byte( REG_ADD_EXT_SENS_DATA_00+i)

    self._write_byte( REG_ADD_REG_BANK_SEL, REG_VAL_REG_BANK_3) #swtich bank3
    
    u8Temp = self._read_byte(REG_ADD_I2C_SLV0_CTRL)
    u8Temp &= ~((REG_VAL_BIT_I2C_MST_EN)&(REG_VAL_BIT_MASK_LEN))
    self._write_byte( REG_ADD_I2C_SLV0_CTRL,  u8Temp)
    
    self._write_byte( REG_ADD_REG_BANK_SEL, REG_VAL_REG_BANK_0) #swtich bank0
	
  def WriteSecondary(self,u8I2CAddr,u8RegAddr,u8data):
    u8Temp=0
    self._write_byte( REG_ADD_REG_BANK_SEL,  REG_VAL_REG_BANK_3) #swtich bank3
    self._write_byte( REG_ADD_I2C_SLV1_ADDR, u8I2CAddr)
    self._write_byte( REG_ADD_I2C_SLV1_REG,  u8RegAddr)
    self._write_byte( REG_ADD_I2C_SLV1_DO,   u8data)
    self._write_byte( REG_ADD_I2C_SLV1_CTRL, REG_VAL_BIT_SLV0_EN|1)

    self._write_byte( REG_ADD_REG_BANK_SEL, REG_VAL_REG_BANK_0) #swtich bank0

    u8Temp = self._read_byte(REG_ADD_USER_CTRL)
    u8Temp |= REG_VAL_BIT_I2C_MST_EN
    self._write_byte( REG_ADD_USER_CTRL, u8Temp)
    utime.sleep(0.01)
    u8Temp &= ~REG_VAL_BIT_I2C_MST_EN
    self._write_byte( REG_ADD_USER_CTRL, u8Temp)

    self._write_byte( REG_ADD_REG_BANK_SEL, REG_VAL_REG_BANK_3) #swtich bank3

    u8Temp = self._read_byte(REG_ADD_I2C_SLV0_CTRL)
    u8Temp &= ~((REG_VAL_BIT_I2C_MST_EN)&(REG_VAL_BIT_MASK_LEN))
    self._write_byte( REG_ADD_I2C_SLV0_CTRL,  u8Temp)

    self._write_byte( REG_ADD_REG_BANK_SEL, REG_VAL_REG_BANK_0) #swtich bank0
	
  def GyroOffset(self):
    s32TempGx = 0
    s32TempGy = 0
    s32TempGz = 0
    for i in range(0,32):
      self.Gyro_Accel_Read()
      s32TempGx += Gyro[0]
      s32TempGy += Gyro[1]
      s32TempGz += Gyro[2]
      utime.sleep(0.01)
    GyroOffset[0] = s32TempGx >> 5
    GyroOffset[1] = s32TempGy >> 5
    GyroOffset[2] = s32TempGz >> 5
	
  def _read_byte(self,cmd):
    rdate = self._bus.readfrom_mem(int(self._address), int(cmd), 1)
    return rdate[0]
  def _read_block(self, reg, length=1):
    return self._bus.readfrom_mem(self._address, reg, length)
	
  def _read_u16(self,cmd):
    LSB = self._bus.readfrom_mem(self._address,cmd, 1)
    MSB = self._bus.readfrom_mem(self._address,cmd+1, 1)
    return (MSB	<< 8) + LSB
	
  def _write_byte(self,cmd,val):
    self._bus.writeto_mem(int(self._address), int(cmd), bytes([int(val)]))
    utime.sleep(0.0001)
	
  def imuAHRSupdata(self,gx, gy,gz,ax,ay,az,mx,my,mz):    
    norm=0.0
    hx = hy = hz = bx = bz = 0.0
    vx = vy = vz = wx = wy = wz = 0.0
    exInt = eyInt = ezInt = 0.0
    ex=ey=ez=0.0 
    halfT = 0.024
    global q0
    global q1
    global q2
    global q3
    q0q0 = q0 * q0
    q0q1 = q0 * q1
    q0q2 = q0 * q2
    q0q3 = q0 * q3
    q1q1 = q1 * q1
    q1q2 = q1 * q2
    q1q3 = q1 * q3
    q2q2 = q2 * q2   
    q2q3 = q2 * q3
    q3q3 = q3 * q3          

    norm = float(1/math.sqrt(ax * ax + ay * ay + az * az))     
    ax = ax * norm
    ay = ay * norm
    az = az * norm

    norm = float(1/math.sqrt(mx * mx + my * my + mz * mz))      
    mx = mx * norm
    my = my * norm
    mz = mz * norm

    # compute reference direction of flux
    hx = 2 * mx * (0.5 - q2q2 - q3q3) + 2 * my * (q1q2 - q0q3) + 2 * mz * (q1q3 + q0q2)
    hy = 2 * mx * (q1q2 + q0q3) + 2 * my * (0.5 - q1q1 - q3q3) + 2 * mz * (q2q3 - q0q1)
    hz = 2 * mx * (q1q3 - q0q2) + 2 * my * (q2q3 + q0q1) + 2 * mz * (0.5 - q1q1 - q2q2)         
    bx = math.sqrt((hx * hx) + (hy * hy))
    bz = hz     

    # estimated direction of gravity and flux (v and w)
    vx = 2 * (q1q3 - q0q2)
    vy = 2 * (q0q1 + q2q3)
    vz = q0q0 - q1q1 - q2q2 + q3q3
    wx = 2 * bx * (0.5 - q2q2 - q3q3) + 2 * bz * (q1q3 - q0q2)
    wy = 2 * bx * (q1q2 - q0q3) + 2 * bz * (q0q1 + q2q3)
    wz = 2 * bx * (q0q2 + q1q3) + 2 * bz * (0.5 - q1q1 - q2q2)  

    # error is sum of cross product between reference direction of fields and direction measured by sensors
    ex = (ay * vz - az * vy) + (my * wz - mz * wy)
    ey = (az * vx - ax * vz) + (mz * wx - mx * wz)
    ez = (ax * vy - ay * vx) + (mx * wy - my * wx)

    if (ex != 0.0 and ey != 0.0 and ez != 0.0):
      exInt = exInt + ex * Ki * halfT
      eyInt = eyInt + ey * Ki * halfT  
      ezInt = ezInt + ez * Ki * halfT

      gx = gx + Kp * ex + exInt
      gy = gy + Kp * ey + eyInt
      gz = gz + Kp * ez + ezInt

    q0 = q0 + (-q1 * gx - q2 * gy - q3 * gz) * halfT
    q1 = q1 + (q0 * gx + q2 * gz - q3 * gy) * halfT
    q2 = q2 + (q0 * gy - q1 * gz + q3 * gx) * halfT
    q3 = q3 + (q0 * gz + q1 * gy - q2 * gx) * halfT  

    norm = float(1/math.sqrt(q0 * q0 + q1 * q1 + q2 * q2 + q3 * q3))
    q0 = q0 * norm
    q1 = q1 * norm
    q2 = q2 * norm
    q3 = q3 * norm
	
  def icm20948Check(self):
    bRet=false
    if REG_VAL_WIA == self._read_byte(REG_ADD_WIA):
      bRet = true
    return bRet
	
  def MagCheck(self):
    self.icm20948ReadSecondary( I2C_ADD_ICM20948_AK09916|I2C_ADD_ICM20948_AK09916_READ,REG_ADD_MAG_WIA1, 2)
    if (pu8data[0] == REG_VAL_MAG_WIA1) and ( pu8data[1] == REG_VAL_MAG_WIA2) :
        bRet = true
        return bRet
		
  def CalAvgValue(self):
    MotionVal = [0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0]
    MotionVal[0]=Gyro[0]/32.8
    MotionVal[1]=Gyro[1]/32.8
    MotionVal[2]=Gyro[2]/32.8
    MotionVal[3]=Accel[0]
    MotionVal[4]=Accel[1]
    MotionVal[5]=Accel[2]
    MotionVal[6]=Mag[0]
    MotionVal[7]=Mag[1]
    MotionVal[8]=Mag[2]
    return MotionVal


  def getdata(self):
    self.Gyro_Accel_Read()
    self.MagRead()
    
    MotionVal=[0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0]
    MotionVal = self.CalAvgValue()
    utime.sleep(0.1)
    self.imuAHRSupdata(MotionVal[0] * 0.0175, MotionVal[1] * 0.0175,MotionVal[2] * 0.0175,
        MotionVal[3],MotionVal[4],MotionVal[5], 
        MotionVal[6], MotionVal[7], MotionVal[8])
    
    pitch = math.asin(-2 * q1 * q3 + 2 * q0* q2)* 57.3
    roll  = math.atan2(2 * q2 * q3 + 2 * q0 * q1, -2 * q1 * q1 - 2 * q2* q2 + 1)* 57.3
    yaw   = math.atan2(-2 * q1 * q2 - 2 * q0 * q3, 2 * q2 * q2 + 2 * q3 * q3 - 1) * 57.3
    
    return [roll,pitch,yaw, Accel[0],Accel[1],Accel[2], Gyro[0],Gyro[1],Gyro[2],Mag[0],Mag[1],Mag[2]]
# ========End of 9-DOF sensor Driver ===========

# ======== Main ==========
pwm = PWM(Pin(BL))
pwm.freq(1000)
pwm.duty_u16(32768)#max 65535
# === Splash =========
LCD = LCD_1inch8()
LCD.fill(0x0)
LCD.rect(0,00,160,128,colour(0,0,255)) # Blue Frame
centrestring("Raspberry Pi Pico",5,1,0,0,colour(255,0,0))
centrestring("Enviro",20,3,0,0,colour(255,255,0))
centrestring("Sensing",45,3,0,0,colour(255,255,0))
#centrestring("with",73,1,0,0,colour(200,200,200))
centrestring("Workout",75,2,0,0,colour(255,0,0))
centrestring("Tony Goodhew",100,1,0,0,colour(0,255,0))
LCD.show()
LCD.text("WS Enviro Sensor", 18,10,colour(255,0,0))               # Uses font from framebuf
utime.sleep(5)
# ========= Detail ===========
LCD.fill(0x0)
LCD.rect(0,0,160,128,colour(0,0,255)) # Blue Frame
LCD.text("WS Enviro Sensor", 5,10,colour(255,255,0))               # Uses font from framebuf
LCD.text("* BME280  T H P",5,30,colour(255,255,255))
LCD.text("* SGP40   Ethanol",5,50,colour(255,255,255))
LCD.text("* TSL2591 Light",5,70,colour(255,255,255))
LCD.text("* LTR390  U.Violet",5,90,colour(255,255,255))
LCD.text("= Taking Readings =",3  ,110,colour(200,200,200))
LCD.show()
LCD.fill(0x0)
LCD.rect(0,0,160,128,colour(0,0,255)) # Blue Frame
LCD.text("WS Enviro Sensor", 18,10,colour(255,0,0))
# ===== Setup sensors ===========
sensor280 = BME280()
sensor280.get_calib_param()

sgp = SGP40()

Luxsensor = TSL2591()
Luxsensor.SET_LuxInterrupt(20, 200)

UVsensor = LTR390()
UVS = UVsensor.UVS()

icm20948 = ICM20948()
icm = []
utime.sleep(1)
# ==== Get readings ===========
for loops in range(2):
    LCD.fill(0x0)
    LCD.rect(0,0,160,128,colour(0,0,255)) # Blue Frame
    LCD.text("WS Enviro Sensor", 18,10,colour(255,0,0))
# BME280    
    pres,temp,hum = sensor280.readData()
#    print(str(temp)+" "+str(hum)+" "+str(pres))
    T = "Temp. C : " + str(temp)
    H = "Hum   % : " + str(hum)    
    P = "Pres hPa: " + str(pres)
    print(T)
    print(H)
    print(P)
    LCD.text("BME280",35,20,colour(255,255,255))
    LCD.text(T, 8,30,colour(0,255,0))
    LCD.text(H, 8,40,colour(255,255,0))
    LCD.text(P, 8,50,colour(0,255,255))
# SGP40 - Gas - Ambient air Quality - VOC - ethanol equivalent   
    print("Raw Gas: ", sgp.measureRaw(25, 50))
    LCD.text("SGP40",35,60,colour(255,255,255))
    LCD.text("Raw Gas : " + str( sgp.measureRaw(25, 50)), 8,70,colour(0,255,0))
# TSL2591 - Light - Lux   
    lux, ir = Luxsensor.Lux() # I've added the IR to the returns
    print("Lux: %d" %lux)
    print("IR: %d" %ir)       # Not sure if this is right or units ??
    LCD.text("TSL2591",35,80,colour(255,255,255))
    LCD.text("Lux: " + str(int(lux)) +"   IR: "+str(ir), 8,90,colour(0,255,0))
# LTR390 - Ultra Violet    
    print("UVS: %d" %UVS)
    LCD.text("LTR390",35,100,colour(255,255,255))
    LCD.text("UV      : " + str(int(UVS)), 8,110,colour(0,255,0))
    LCD.show()
# ICM20948 - 9-DoF sensor
    icm = icm20948.getdata()
    print("Roll = %.2f , Pitch = %.2f , Yaw = %.2f" %(icm[0],icm[1],icm[2]))
    print("Acceleration: X = %d, Y = %d, Z = %d" %(icm[3],icm[4],icm[5]))
    print("Gyroscope:     X = %d , Y = %d , Z = %d" %(icm[6],icm[7],icm[8]))
    print("Magnetic:      X = %d , Y = %d , Z = %d" %(icm[9],icm[10],icm[11]))
    print(" ")
    utime.sleep(2.5)
    LCD.fill(0x0)
    LCD.text("ICM20948 9-DoF",20,20,colour(255,255,255))
    LCD.text("Roll:  " + str(icm[0]), 8,30,colour(0,255,0))
    LCD.text("Pitch: " + str(icm[1]), 8,40,colour(255,255,0))           
    LCD.text("Yaw:   " + str(icm[2]), 8,50,colour(0,255,255))
    LCD.text("Accelerometer",28,70,colour(255,255,255))
    LCD.text("X: " + str(icm[3]), 28,80,colour(0,255,0))
    LCD.text("Y: " + str(icm[4]), 28,90,colour(255,255,0))           
    LCD.text("Z: " + str(icm[5]), 28,100,colour(0,255,255))
    LCD.show()
    utime.sleep(2.5)
    LCD.fill(0x0)
    LCD.text("Gyroscope",28,20,colour(255,255,255))
    LCD.text("Roll:  " + str(icm[6]), 28,30,colour(0,255,0))
    LCD.text("Pitch: " + str(icm[7]), 28,40,colour(255,255,0))           
    LCD.text("Yaw:   " + str(icm[8]), 28,50,colour(0,255,255))
    LCD.text("Magnetometer",28,70,colour(255,255,255))
    LCD.text("X: " + str(icm[9]), 28,80,colour(0,255,0))
    LCD.text("Y: " + str(icm[10]), 28,90,colour(255,255,0))           
    LCD.text("Z: " + str(icm[11]), 28,100,colour(0,255,255))
    LCD.show()
    utime.sleep(2.5)
LCD.fill(0x0)
LCD.show()

